#include <stdio.h>

#ifdef __DJGPP__
#undef REALTIME
#include <inlines/pc.h>
#include <dos.h>
#else
#ifndef _OS2_
#include <getopt.h>
#  if defined __GLIBC__ && __GLIBC__ >= 2
#    include <sys/io.h>
#  else
#    ifdef _AXP_
#       include <sys/io.h>
#    else
#       include <asm/io.h>
#    endif
#  endif												   /* __GLIBC__ */
#else
#include <sys/hw.h>
#endif /* _OS2_ */
#endif

#include <unistd.h>
#include <stdlib.h>
#include <string.h>										   /* For strncpy */

#define TAG_DEBUG 0
#define DUMPADDR 0

#include "pnp.h"
#include "resource.h"

#define NUM_CARDS 128
#define IDENT_LEN 9
#define TMP_LEN 16
#define LARGE_LEN 65536

static char rcsid[] = "$Id: pnpdump.c,v 1.18 1999/02/14 22:47:18 fox Exp $";

#ifdef REALTIME
/*
 * Code to set scheduler to Realtime, Round-Robin, so usleeps right etc
 */
#if defined __GLIBC__ && __GLIBC__ >= 2
#  include <sched.h>
#  include <time.h>
#else
#  include <linux/unistd.h>
#  include <linux/sched.h>
#  include <sys/time.h>

#  ifdef NEEDSETSCHEDULER
_syscall3(int, sched_setscheduler, pid_t, pid, int, policy, struct sched_param *, sched_p)
#  endif
#  ifdef NEEDNANOSLEEP
_syscall2(int, nanosleep, struct timespec *, rqtp, struct timespec *, rmtp)
#  endif
#endif /* __GLIBC__ */

int realtimeok = 0;

void
setroundrobin(void)
{
	pid_t mypid = getpid();
	struct sched_param sched_p =
	{50};
	/* Use the syscall, as my library not up todate */
	if (sched_setscheduler(mypid, SCHED_RR, &sched_p) < 0)
	{
		perror("Couldn't set real-time scheduling, may be a bit slow");
	}
	else
		realtimeok = 1;
}

/*
 * Use nanosleep for realtime delays
 */
void
delaynus(long del)
{
	struct timespec t;
	t.tv_sec = 0;
	t.tv_nsec = del * 1000;
	/*
	 * Need to handle case where binary for later kernel run on an earlier
	 * one, which doesn't support nanosleep (emergency backup spare !)
	 */
	if (realtimeok)
	{
		if (nanosleep(&t, (struct timespec *) 0) < 0)
		{
			perror("nanosleep failed");
			realtimeok = 0;
		}
	}
	else
		usleep(del);
}

#else
#define delaynus(x) usleep(x)
#undef NEEDSETSCHEDULER
#undef NEEDNANOSLEEP
#endif /* REALTIME */

#ifdef _OS2_
unsigned 
inb(unsigned port)
{
	return (_inp8(port));
}

unsigned 
inw(unsigned port)
{
	return (_inp16(port));
}

unsigned 
inl(unsigned port)
{
	return (_inp32(port));
}

void 
outb(unsigned value, unsigned port)
{
	_outp8(port, value);
}

/* void outw(unsigned port, unsigned value){_outp16(port, value);} */
/* void outl(unsigned port, unsigned value){_outp32(port, value);} */
unsigned 
iopl(unsigned port)
{
	return (0);
}

unsigned 
usleep(unsigned wait)
{
	return (_sleep2((wait + 990) / 1000));
}

#endif /* _OS2_ */

#ifdef __DJGPP__
unsigned 
iopl(unsigned port)
{
	return (0);
}

unsigned 
usleep(unsigned wait)
{
	return (delay((wait + 990) / 1000), 0);
}

#endif /* __DJGPP__ */

#define ASSERT(condition)								   /* as nothing */

typedef void (*resource_handler)
     (FILE * output_file, struct resource * res, int selected);

static void read_board_resources(struct resource *first_element,
								 int dependent_clause,
								 struct resource **result_array_ptr,
								 int *count);

static void print_resource(FILE * output_file,
						   struct resource *res,
						   int selected);

static void for_each_resource(struct resource *res,
							  int num_resources,
							  resource_handler handler,
							  FILE * output_file,
							  int selected);

#ifndef __DJGPP__
int do_script;
char *output_script = NULL;

static void print_script_header(FILE * script_file);

static void print_script(FILE * output_file,
						 struct resource *res,
						 int selected);
#endif /* __DJGPP__ */

static void read_board(int csn, struct resource **res_list, int *list_len);

static void unread_resource_data(char prev);

/* Global state variables used for printing resource information. */
unsigned long serialno = 0;
char devid[8];											   /* Copy as devidstr
														    * returns pointer to
														    * static buffer subject
														    * to change */
int nestdepth;
int curid;
int curdma, depdma;
int curio, depio;
int curint, depint;
int curmem, depmem;
int starteddeps;
char *indep = "";

/* This shouldn't be required 8-Oct-1998 */
static void
prepare_to_print(void)
{
	nestdepth = 0;
	curid = 0;
	curdma = 0;
	depdma = 0;
	curio = 0;
	depio = 0;
	curint = 0;
	depint = 0;
	curmem = 0;
	depmem = 0;
	starteddeps = 0;
	indep = "";
}

unsigned char serial_identifier[NUM_CARDS + 1][IDENT_LEN];
unsigned char csumdiff[NUM_CARDS + 1];

unsigned char tmp[TMP_LEN];
unsigned char large[LARGE_LEN];
static char *devstring = 0;

int showmasks = 0;
int do_autoconfig = 0;
int do_fullreset = 0;

int csum = 0;
int backspaced = 0;
char backspaced_char;
int read_port = MIN_READ_ADDR;
int boards_found = 0;

int do_isolate(void);
void initiate(void);
void dumpdata(int);
void read_idents(void);

static void usage(char *program_name);

int do_dumpregs = 0;
int do_ignorecsum = 0;
void dumpregs(int);

char *devidstr(unsigned char, unsigned char, unsigned char, unsigned char);

/*
 * Single argument is initial readport
 * Two arguments is no-of-boards and the readport to use
 */
int
main(int argc, char **argv)
{
	int i;
	char *program = argv[0];
	struct resource *resources;
	int resource_count;
	int alloc_result;
#ifndef __DJGPP__
	const struct option longopts[] =
	{
		{"masks", no_argument, NULL, 'm'},
		{"config", no_argument, NULL, 'c'},
		{"script", optional_argument, NULL, 's'},
		{"reset", no_argument, NULL, 'r'},
		{"dumpregs", no_argument, NULL, 'd'},
		{"ignorecsum", no_argument, NULL, 'i'},
		{0, 0, 0, 0},
	};
	int opt;

	while ((opt = getopt_long(argc, argv, "cms:rdi", longopts, NULL)) != EOF)
	{
		switch (opt)
		{
		case 'c':
			do_autoconfig = 1;
			break;
		case 'm':
			showmasks = 1;
			break;
		case 'r':
			do_fullreset = 1;
			break;
		case 's':
			do_script = 1;
			output_script = optarg;
			break;
		case 'd':
			do_dumpregs = 1;
			break;
		case 'i':
			do_ignorecsum = 1;
			break;
		case '?':
#if 0
			fprintf(stderr, "unrecognized option.\n");
#endif
			usage(program);
			return 1;
		case ':':
#if 0
			fprintf(stderr, "missing parameter.\n");
#endif
			usage(program);
			return 1;
		default:
			fprintf(stderr,
					"?? getopt returned character code 0x%x ('%c').\n",
					opt, opt);
			return 1;
		}

	}
	argc -= optind - 1;
	argv += optind - 1;
#else
	while (argc > 1)
	{
		if (strcmp(argv[1], "--reset") == 0)
		{
			do_fullreset = 1;
		}
		else if (strcmp(argv[1], "--masks") == 0)
		{
			showmasks = 1;
		}
		else if (strcmp(argv[1], "--dumpregs") == 0)
		{
			do_dumpregs = 1;
		}
		else if (strcmp(argv[1], "--config") == 0)
		{
			do_autoconfig = 1;
		}
		else if (strcmp(argv[1], "--ignorecsum") == 0)
		{
			do_ignorecsum = 1;
		}
		else
		{
			break;
		}
		argv++;
		argc--;
	}
#endif /* __DJGPP__ */

	if ((argc < 1) || (argc > 3))
	{
		usage(program);
		return 1;
	}

	/* If a number of boards and read_port have been given, don't ISOLATE */
	if (argc == 3)
	{
		boards_found = atoi(argv[1]);	/* This stops ISOLATE */
		if ((boards_found < 0) || (boards_found >= NUM_CARDS))
		{
			fprintf(stderr, "Cannot handle %d boards, recompile with NUM_CARDS bigger\n", boards_found);
			exit(1);
		}
	}
	if (argc > 1)
	{
		/* Read decimal or hex number */
		read_port = (int) strtol(argv[argc - 1], (char **) NULL, 0);
		if ((read_port < MIN_READ_ADDR) || (read_port > MAX_READ_ADDR))
		{
			fprintf(stderr, "Port address %s (0x%04x) out of range 0x203..0x3ff\n", argv[argc - 1], read_port);
			exit(1);
		}
		read_port |= 3;
	}
#ifdef _AXP_
	/* ALPHA only has ioperm, apparently, so cover all with one permission */
	if (ioperm(MIN_READ_ADDR, WRITEDATA_ADDR - MIN_READ_ADDR + 1, 1))
#else
	/*
	 * Have to get unrestricted access to io ports, as WRITE_DATA port >
	 * 0x3ff
	 */
	if (iopl(3) < 0)
#endif
	{
		perror("Unable to get io permission for WRITE_DATA");
		exit(1);
	}
#ifdef REALTIME
	setroundrobin();
#endif /* REALTIME */
	printf("# %s\n# This is free software, see the sources for details.\n"
		   "# This software has NO WARRANTY, use at your OWN RISK\n"
		   "#\n"
		   "# For details of this file format, see isapnp.conf(5)\n"
		   "#\n"
		   "# For latest information and FAQ on isapnp and pnpdump see:\n"
		   "# http://www.roestock.demon.co.uk/isapnptools/\n"
		   "#\n"
		   "# Compiler flags:", rcsid);
#ifdef _AXP_
	printf(" -D_AXP_");
#endif /* _AXP_ */
#ifdef REALTIME
	printf(" -DREALTIME");
#endif /* REALTIME */
#ifdef NEEDSETSCHEDULER
	printf(" -DNEEDSETSCHEDULER");
#endif /* NEEDSETSCHEDULER */
#ifdef NEEDNANOSLEEP
	printf(" -DNEEDNANOSLEEP");
#endif /* NEEDNANOSLEEP */
#ifdef ABORT_ONRESERR
	printf(" -DABORT_ONRESERR");
#endif /* ABORT_ONRESERR */
	printf("\n#\n");
	/* Have to do this anyway to check readport for clashes */
	alloc_system_resources();
	/* Now get board info */
	initiate();
	resources = NULL;
	resource_count = 0;
	for (i = 1; i <= boards_found; i++)
	{
#if 1
		read_board(i, &resources, &resource_count);
#else
		dumpdata(i);
#endif
		if(do_dumpregs)
			dumpregs(i);
	}
	if (do_autoconfig)
	{
		alloc_result = alloc_resources(resources, resource_count, NULL, 0);
	}
	else
	{
		alloc_result = 0;
	}
	prepare_to_print();
	for_each_resource(resources, resource_count,
					  print_resource, stdout, alloc_result);
	/* Musn't forget this ! */
	printf("# Returns all cards to the \"Wait for Key\" state\n");
	printf("(WAITFORKEY)\n");
	/* Release resources */
#ifdef _AXP_
	ioperm(MIN_READ_ADDR, WRITEDATA_ADDR - MIN_READ_ADDR + 1, 0);
#else
	(void) iopl(0);
#endif

#ifndef __DJGPP__
	if (do_script)
	{
		FILE *script_file;

		if (output_script != NULL)
		{
			script_file = fopen(output_script, "w");
		}
		else
		{
			script_file = popen("sh", "w");
		}
		if (script_file == NULL)
		{
			perror(output_script ? output_script : "sh");
			exit(1);
		}
		if (alloc_result == 0)
		{
			fprintf(script_file, "#!/bin/sh\n" "isa-pnp-config-failed\n");
		}
		else
		{
			print_script_header(script_file);
			for_each_resource(resources, resource_count,
							  print_script, script_file, 1);
		}
		if (output_script)
		{
			fclose(script_file);
		}
		else
		{
			pclose(script_file);
		}
	}
#endif /* __DJGPP__ */
	return (0);
}

static void 
usage(char *program_name)
{
	fprintf(stderr, "\n%s\nThis is free software, see the sources for details.\n"
			"For latest information, see "
			"http://www.roestock.demon.co.uk/isapnptools/\n"
			"This software has NO WARRANTY, use at your OWN RISK\n\n"
#ifndef __DJGPP__
			"Usage: %s [--masks] [--config] [--reset] [--script=outputfile] [--dumpregs] [--ignorecsum]\n",
#else
			"Usage: %s [--masks] [--config] [--reset] [--dumpregs] [--ignorecsum]\n",
#endif
			rcsid, program_name);
}

/*
 * Wait for resource data byte
 *
 * Return true if error occurred (timeout)
 */
int
statuswait(void)
{
#ifdef REALTIME
#define TIMEOUTLOOPS 100
	int to;												   /* For timeout */
	/*
	 * Try for up to 1ms
	 */
	for (to = 0; to < TIMEOUTLOOPS; to++)
	{
		if (STATUS & 1)
			break;
		delaynus(10L);
	}
	if (to >= TIMEOUTLOOPS)
	{
		fprintf(stderr, "Timeout attempting to read resource data - is READPORT correct ?\n");
		return (1);
	}
#else /* !REALTIME */
#if defined _OS2_ || defined __DJGPP__
#define TIMEOUTLOOPS 2
	int to;												   /* For timeout */
	/*
	 * Try for up to 2ms
	 */
	for (to = 0; to < TIMEOUTLOOPS; to++)
	{
		usleep(1000);
		if (STATUS & 1)
			break;
	}
	if (to >= TIMEOUTLOOPS)
	{
		fprintf(stderr, "Timeout attempting to read resource data - is READPORT correct ?\n");
		return (1);
	}
#else /* !(OS2 or DJGPP) */
	/*
	 * Infinite loop potentially, but if we usleep, we may lose 10ms
	 */
	while (!(STATUS & 1))
		;
#endif /* !(OS2 or DJGPP) */
#endif /* !REALTIME */
	return (0);
}

int port;

void
sendkey(void)
{
	static char initdata[INIT_LENGTH] = INITDATA;
	int i;
	ADDRESS(0);
	ADDRESS(0);
	for (i = 0; i < INIT_LENGTH; i++)
	{
		ADDRESS(initdata[i]);
	}
}

void
initiate(void)
{
	int i;
	if (!boards_found)
	{
		/* All cards now isolated, read the first one */
		for (port = read_port; port <= MAX_READ_ADDR; port += READ_ADDR_STEP)
		{
			/* Check it doesn't clash with anything */
			if(!allocate_resource(IOport_TAG, port, 1, "readport"))
			{
				continue;
			}
			deallocate_resource(IOport_TAG, port, 1);
			/* Make sure all cards are in Wait for key */
			CONFIGCONTROL;
			WRITE_DATA(CONFIG_WAIT_FOR_KEY);
			/* Make them listen */
			delaynus(2000L);
			sendkey();
			/* Reset the cards */
			CONFIGCONTROL;
			WRITE_DATA(CONFIG_RESET_CSN);
			if(do_fullreset)
			{
				CONFIGCONTROL;
				WRITE_DATA(CONFIG_RESET);
				delaynus(2000L);
			}
			printf("# Trying port address %04x\n", port);
			if (do_isolate())
				break;
		}
		if (port > MAX_READ_ADDR)
		{
			printf("# No boards found\n");
			exit(0);
		}
		printf("# Board %d has serial identifier", boards_found);
		for (i = IDENT_LEN; i--;)
			printf(" %02x", serial_identifier[boards_found][i]);
		printf("\n");
		while (do_isolate())
		{
			printf("# Board %d has serial identifier", boards_found);
			for (i = IDENT_LEN; i--;)
				printf(" %02x", serial_identifier[boards_found][i]);
			printf("\n");
		}
		printf("\n# (DEBUG)\n(READPORT 0x%04x)\n", port);
		if(do_ignorecsum)
			printf("(IGNORECRC)\n");
		if(do_fullreset)
			printf("(ISOLATE CLEAR)\n");
		else
			printf("(ISOLATE PRESERVE)\n");
		printf("(IDENTIFY *)\n");
	}
	else
	{
		CONFIGCONTROL;
		WRITE_DATA(CONFIG_WAIT_FOR_KEY);
		sendkey();
		read_idents();
		printf("\n# (DEBUG)\n(READPORT 0x%04x)\n(CSN %d)\n(IDENTIFY *)\n", read_port, boards_found);
	}
	printf("(VERBOSITY 2)\n(CONFLICT (IO FATAL)(IRQ FATAL)(DMA FATAL)(MEM FATAL)) # or WARNING\n\n");
}

void
read_idents(void)
{
	int csn = 0;
	int i;
	for (csn = 1; csn <= boards_found; csn++)
	{
		Wake(csn);
		for (i = 0; i < IDENT_LEN; i++)
		{
			if (statuswait())
				return;
			serial_identifier[csn][i] = RESOURCEDATA;
		}
		printf("# Board %d has serial identifier", csn);
		for (i = IDENT_LEN; i--;)
			printf(" %02x", serial_identifier[csn][i]);
		printf("\n");
		if (serial_identifier[csn][0] == 0xff)
			boards_found = csn - 1;
	}
}

int
do_isolate(void)
{
	unsigned char c1, c2;
	int i;
	int index;
	int newbit;
	int goodaddress = 0;
	if (boards_found >= NUM_CARDS)
	{
		fprintf(stderr, "Too many boards found, recompile with NUM_CARDS bigger\n");
		return 0;
	}
	csum = 0x6a;
	/* Assume we will find one */
	boards_found++;
	Wake(0);
	SetRdPort(port);
	delaynus(1000L);
	SERIALISOLATION;
	delaynus(2000L);
	for (index = 0; index < IDENT_LEN - 1; index++)
	{
		for (i = 0; i < 8; i++)
		{
			newbit = 0x00;
			delaynus(250L);
			c1 = READ_DATA;
			delaynus(250L);
			c2 = READ_DATA;
			if (c1 == 0x55)
			{
				if (c2 == 0xAA)
				{
					goodaddress = 1;
					newbit = 0x80;
				}
				else
				{
					goodaddress = 0; 
				}
			}
			/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
			serial_identifier[boards_found][index] >>= 1;
			serial_identifier[boards_found][index] |= newbit;
			/* Update checksum */
			if (((csum >> 1) ^ csum) & 1)
				newbit ^= 0x80;
			csum >>= 1;
			csum |= newbit;
		}
		/* printf(" *** %02x \n", serial_identifier[boards_found][index]); */
	}
	/* printf("computed csum is %02x", csum); */
	for (i = 0; i < 8; i++)
	{
		newbit = 0x00;
		delaynus(250L);
		c1 = READ_DATA;
		delaynus(250L);
		c2 = READ_DATA;
		if (c1 == 0x55)
		{
			if (c2 == 0xAA)
			{
				goodaddress = 1;
				newbit = 0x80;
			}
		}
		/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
		serial_identifier[boards_found][index] >>= 1;
		serial_identifier[boards_found][index] |= newbit;
	}
	/* printf(" *** csum *** %02x\n", serial_identifier[boards_found][index]); */
	if (goodaddress && (do_ignorecsum || (csum == serial_identifier[boards_found][index])))
	{
		CARDSELECTNUMBER;
		WRITE_DATA(boards_found);
		if((csumdiff[boards_found] = (csum - serial_identifier[boards_found][index])))
			printf("# WARNING: serial identifier mismatch for board %d: expected 0x%02X, got 0x%02X\n", boards_found, serial_identifier[boards_found][index], csum);
		return (1);
	}
	/* We didn't find one */
	boards_found--;
	return (0);
}

char *
devidstr(unsigned char d1, unsigned char d2, unsigned char d3, unsigned char d4)
{
	static char resstr[] = "PNP0000";
	sprintf(resstr, "%c%c%c%x%x%x%x", 'A' + (d1 >> 2) - 1, 'A' + (((d1 & 3) << 3) | (d2 >> 5)) - 1,
			'A' + (d2 & 0x1f) - 1, d3 >> 4, d3 & 0x0f, d4 >> 4, d4 & 0x0f);
	return resstr;
}

void
lengtherror(int len, char *msg)
{
	int i;
	printf("# Bad tag length for %s in 0x%02x", msg, tmp[0]);
	if (tmp[0] & 0x80)
	{
		printf(" 0x%02x 0x%02x", len & 255, len >> 8);
		for (i = 0; i < len; i++)
			printf(" 0x%02x", large[i]);
	}
	else
	{
		for (i = 1; i <= len; i++)
			printf(" 0x%02x", tmp[i]);
	}
	printf("\n");
}

#ifdef ABORT_ONRESERR
#define BREAKORCONTINUE   goto oncorruptresourcedata
#else
#define BREAKORCONTINUE   break
#endif

static void
showmask(unsigned long mask)
{
	int i;
	int firsttime = 1;

	if (showmasks)
	{
		printf(" mask 0x%lx", mask);
	}
	else
	{
		for (i = 0; mask; i++, mask >>= 1)
		{
			if (mask & 0x1)
			{
				if (!firsttime)
				{
					printf(mask == 1 ? " or " : ", ");
				}
				else
				{
					printf(" ");
					firsttime = 0;
				}
				printf("%d", i);
			}
		}
	}
}

static void 
unread_resource_data(char prev)
{
	csum = (csum - prev) & 0xFF;
	backspaced = 1;
	backspaced_char = prev;
}

/* returns 0 on success, nonzero on failure. */

static inline int
read_resource_data(unsigned char *result)
{
	if (backspaced)
	{
		*result = backspaced_char;
		backspaced = 0;
		return 0;
	}
	if (statuswait())
		return 1;
	*result = RESOURCEDATA;
	csum = (csum + *result) & 0xFF;
	return 0;
}

static int
read_one_resource(struct resource *res)
{
	int i;
	if (read_resource_data(&res->tag) != 0)
		return 1;
	if (res->tag & 0x80)
	{
		/* Large item */
		res->type = res->tag;
		for (i = 1; i <= 2; i++)
		{
			if (read_resource_data(&tmp[i]) != 0)
				return 1;
		}
		res->len = (tmp[2] << 8) | tmp[1];
	}
	else
	{
		/* Small item */
		res->type = (res->tag >> 3) & 0x0f;
		res->len = res->tag & 7;
	}
	if ((res->data = malloc(res->len * sizeof(unsigned char))) == NULL)
	{
		fprintf(stderr, "read_one_resource: memory allocation failure.\n");
		exit(1);
	}
	for (i = 0; i < res->len; i++)
	{
		if (read_resource_data(&res->data[i]) != 0)
			return 1;
	}

#if TAG_DEBUG
	printf("# %s %02X Tag %02X[%02X] :",
		   (res->tag & 0x80) ? "L" : "S", res->tag, res->type, res->len);
	for (i = 0; i <= res->len; i++)
	{
		printf(" %02x", res->data[i]);
	}
	printf("\n");
#endif
	return 0;
}

static void
fill_in_resource(struct resource *res)
{
	res->mask = ~0;
	res->start = 0;				/* Set some defaults */
	res->size = 1;
	res->step = 1;
	switch (res->type)
	{
	case IRQ_TAG:
		res->end = 16;
		res->mask = res->data[0] | (res->data[1] << 8);
		break;
	case DMA_TAG:
		res->end = 8;
		res->mask = res->data[0];
		break;
	case IOport_TAG:
		res->start = (res->data[2] << 8) | res->data[1];
		res->end = ((res->data[4] << 8) | res->data[3]) + 1;
		res->step = res->data[5];
		res->size = res->data[6];
		break;
	case FixedIO_TAG:
		res->start = (res->data[1] << 8) | res->data[0];
		res->size = res->data[2];
		res->end = res->start + 1;
		break;
	case MemRange_TAG:
		res->start = (((long) res->data[2] << 8) | res->data[1]) * 256;
		res->end = (((long) res->data[4] << 8) | res->data[3]) * 256 + 1;
		res->step = (res->data[6] << 8) | res->data[5];
		res->size = (((long) res->data[8] << 8) | res->data[7]) * 256;
		break;
	case Mem32Range_TAG:
		res->start = res->end = 0;
		/* STUB */
		break;
	case FixedMem32Range_TAG:
		res->start = res->end = 0;
		/* STUB */
		break;
	default:
		break;
	}
	/* Set initial value. */
	for (res->value = res->start; res->value < res->end; res->value += res->step)
	{
		if ((1 << (res->value & 0x1F)) & res->mask)
			break;
	}
}

static void
read_alternative(struct alternative *alt, struct resource *start_element)
{
	ASSERT(start_element->type == StartDep_TAG);
	switch (start_element->len)
	{
	case 0:
		alt->priority = 1;
		break;					/* 1 == "acceptable" */
	case 1:
		alt->priority = start_element->data[0];
		break;
	default:
		fprintf(stderr, "read_alternative: Illegal StartDep_TAG length %d.\n", start_element->len);
		break;
	}
	alt->resources = NULL;
	alt->len = 0;
	read_board_resources(start_element, 1, &alt->resources, &alt->len);
	*start_element = alt->resources[alt->len - 1];
	(alt->len)--;
}

static void
read_board_resources(struct resource *first_element,
					 int dependent_clause,
					 struct resource **result_array_ptr,
					 int *count)
{
	struct resource *result_array;
	int elts;

	result_array = *result_array_ptr;
	elts = *count;
	if (!result_array)
	{
		result_array = malloc(sizeof(struct resource));
		*count = elts = 0;
	}
	if (!result_array)
	{
		fprintf(stderr,
				"read_board_resources: malloc of result_array failed.\n");
		exit(1);
	}
	if (first_element != NULL)
	{
		elts++;
		if ((result_array =
			 realloc(result_array, elts * sizeof(struct resource))) == NULL)
		{
			fprintf(stderr, "read_board_resources: realloc of result_array failed.\n");
			exit(1);
		}
		result_array[elts - 1] = *first_element;
	}
	do
	{
		elts++;
		result_array = realloc(result_array, elts * sizeof(struct resource));
		if (!result_array)
		{
			fprintf(stderr,
				 "read_board_resources: realloc of result_array failed.\n");
			exit(1);
		}
		memset(tmp, 0, sizeof(tmp));
		read_one_resource(&result_array[elts - 1]);
		if (result_array[elts - 1].type == StartDep_TAG && !dependent_clause)
		{
			int num_alts = 1;
			struct alternative *alts;
			alts = malloc(sizeof(struct alternative));
			do
			{
				if (num_alts != 1)
				{
					alts = realloc(alts, num_alts * sizeof(struct alternative));
				}
				if (!alts)
				{
					fprintf(stderr, "read_board_resources: memory allocation for alternatives failed.\n");
					exit(1);
				}
				read_alternative(&alts[num_alts - 1], &result_array[elts - 1]);
				num_alts++;
			}
			while (result_array[elts - 1].type == StartDep_TAG);
			result_array[elts - 1].start = 0;
			result_array[elts - 1].end = num_alts - 1;
			result_array[elts - 1].mask = ~0;
			result_array[elts - 1].step = 1;
			result_array[elts - 1].alternatives = alts;
		}
		else
		{
			fill_in_resource(&result_array[elts - 1]);
		}
	}
	while (result_array[elts - 1].type != End_TAG &&
		   (!dependent_clause ||
			(result_array[elts - 1].type != EndDep_TAG &&
			 result_array[elts - 1].type != StartDep_TAG)));
	*count = elts;
	*result_array_ptr = result_array;
}

static int IORangeCheck = 0;

static void
print_resource(FILE * output_file, struct resource *res, int selected)
{
	static int logdevno = 0;
	int type = res->type;
	int len = res->len;
	int i;
	char *comment_if_not_selected = selected ? " " : "#";

	switch (type)
	{
	case NewBoard_PSEUDOTAG:
		{
			int csn = res->start;

			logdevno = 0;
			serialno = (unsigned long)
				serial_identifier[csn][4] +
				(serial_identifier[csn][5] << 8) +
				(serial_identifier[csn][6] << 16) +
				(serial_identifier[csn][7] << 24);
			strncpy(devid, devidstr(serial_identifier[csn][0],
									serial_identifier[csn][1],
									serial_identifier[csn][2],
									serial_identifier[csn][3]), 8);

			printf("# Card %d: (serial identifier", csn);
			for (i = IDENT_LEN; i--;)
				printf(" %02x", serial_identifier[csn][i]);
			printf(")\n");
			if (serialno == 0xffffffffUL)
			{
				printf("# Vendor Id %s, No Serial Number (-1), checksum 0x%02X.\n", devid, serial_identifier[csn][8]);
			}
			else
			{
				printf("# Vendor Id %s, Serial Number %lu, checksum 0x%02X.\n", devid, serialno, serial_identifier[csn][8]);
			}
			if (res->len != IDENT_LEN)
			{
				i = 0;
				printf("# Ident byte %d, (%02x) differs from resource data (%02x)\n", i, serial_identifier[csn][i], res->data[0]);
				printf("#Assuming the card is broken and this is the start of the resource data\n");
			}
			else
			{
				for (i = 1; i < IDENT_LEN; i++)
				{
					if ((res->data[i] != serial_identifier[csn][i]) && (i < (IDENT_LEN - 1)))
						printf("# Ident byte %d, (%02x) differs from resource data (%02x)\n", i, serial_identifier[csn][i], res->data[i]);
				}
			}
		}

		break;
	case PnPVerNo_TAG:
		{
			if (len != 2)
			{
				lengtherror(len, "PnPVerNo_TAG");
				BREAKORCONTINUE;
			}
			printf("# %sVersion %d.%d, Vendor version %x.%x\n", indep,
				   res->data[0] >> 4,
				   res->data[0] & 0x0f,
				   res->data[1] >> 4,
				   res->data[1] & 0x0f);
			break;
		}
	case LogDevId_TAG:
		{
			int reg;
			static char *regfns[8] =
			{"Device capable of taking part in boot process",
			 "Device support I/O range check register",
			 "Device supports reserved register @ 0x32",
			 "Device supports reserved register @ 0x33",
			 "Device supports reserved register @ 0x34",
			 "Device supports reserved register @ 0x35",
			 "Device supports reserved register @ 0x36",
			 "Device supports reserved register @ 0x37"};
			if ((len < 5) || (len > 6))
			{
				lengtherror(len, "LogDevId_TAG");
				BREAKORCONTINUE;
			}
			indep = "";
			if (nestdepth)
			{
				/* If we have a device name, show it (the last description string before the END DF flag is it) */
				if (serialno == 0xffffffffUL)
					printf(" (NAME \"%s/-1[%d]", devid, logdevno-1);
				else
					printf(" (NAME \"%s/%lu[%d]", devid, serialno, logdevno-1);
				if(devstring)
					printf("{%-20s}", devstring);
				printf("\")\n");
				printf("%s (ACT Y)\n", comment_if_not_selected);
				while (nestdepth)
				{
					printf(")");
					nestdepth--;
				}
				printf("\n");
			}
			printf("#\n# %sLogical device id %s\n", indep, devidstr(res->data[0], res->data[1], res->data[2], res->data[3]));
			indep = "    ";
			IORangeCheck = 0;
			for (i = 0, reg = 1; reg < 256; i++, reg <<= 1)
			{
				if (res->data[4] & reg)
				{
					printf("# %s%s\n", indep, regfns[i]);
					if(i == 1)
						IORangeCheck = 1;
				}
			}
			for (i = 0, reg = 1; reg < 256; i++, reg <<= 1)
			{
				if (res->data[5] & reg)
				{
					printf("# %sDevice supports vendor reserved register @ 0x%02x\n", indep, 0x38 + i);
				}
			}
			printf("#\n# Edit the entries below to uncomment out the configuration required.\n");
			printf("# Note that only the first value of any range is given, this may be changed if required\n");
			printf("# Don't forget to uncomment the activate (ACT Y) when happy\n\n");
			if (serialno == 0xffffffffUL)
				printf("(CONFIGURE %s/-1 (LD %d\n", devid, logdevno++);
			else
				printf("(CONFIGURE %s/%lu (LD %d\n", devid, serialno, logdevno++);
			nestdepth = 2;
			curdma = 0;
			depdma = 0;
			curio = 0;
			depio = 0;
			curint = 0;
			depint = 0;
			curmem = 0;
			depmem = 0;
			starteddeps = 0;
			break;
		}
	case CompatDevId_TAG:
		{
			if (len != 4)
			{
				lengtherror(len, "CompatDevId_TAG");
				BREAKORCONTINUE;
			}
			printf("# %sCompatible device id %s\n", indep, devidstr(res->data[0], res->data[1], res->data[2], res->data[3]));
			break;
		}
	case IRQ_TAG:
		{
			int firstirq = 0;
			char *edge = "+E";
			if ((len < 2) || (len > 3))
			{
				lengtherror(len, "IRQ_TAG");
				BREAKORCONTINUE;
			}
			if ((len >= 2) && (res->data[0] || res->data[1]))
			{
				printf("# %sIRQ", indep);
				showmask(res->mask);
				printf(".\n");
				firstirq = res->value;
				if (len == 3)
				{
					if (res->data[2] & 1)
					{
						printf("# %s%sHigh true, edge sensitive interrupt\n", indep, indep);
						edge = "+E";
					}
					if (res->data[2] & 2)
					{
						printf("# %s%sLow true, edge sensitive interrupt\n", indep, indep);
						edge = "-E";
					}
					if (res->data[2] & 4)
					{
						printf("# %s%sHigh true, level sensitive interrupt\n", indep, indep);
						edge = "+L";
					}
					if (res->data[2] & 8)
					{
						printf("# %s%sLow true, level sensitive interrupt\n", indep, indep);
						edge = "-L";
					}
				}
				else
				{
					printf("# %s%sHigh true, edge sensitive interrupt (by default)\n", indep, indep);
				}
				printf("%s (INT %d (IRQ %ld (MODE %s)))\n",
					   comment_if_not_selected, curint, res->value, edge);
				curint++;
				if (!starteddeps)
					depint = curint;
			}
			else
			{
				printf("# %s*** ERROR *** No IRQ specified!\n", indep);
			}
			break;
		}
	case DMA_TAG:
		{
			int firstdma = 4;							   /* Ie, no DMA */
			if (len != 2)
			{
				lengtherror(len, "DMA_TAG");
				BREAKORCONTINUE;
			}
			if (res->mask)
			{
				printf("# %s%sDMA channel",
					   indep, (curdma == 0) ? "First " : "Next ");
				firstdma = res->value;
				showmask(res->mask);
				printf(".\n");
			}
			else
			{
				printf("# %s*** ERROR *** No DMA channel specified!\n", indep);
			}
			if ((res->data[1] & 3) == 0)
				printf("# %s%s8 bit DMA only\n", indep, indep);
			if ((res->data[1] & 3) == 1)
				printf("# %s%s8 & 16 bit DMA\n", indep, indep);
			if ((res->data[1] & 3) == 2)
				printf("# %s%s16 bit DMA only\n", indep, indep);
			printf("# %s%sLogical device is%s a bus master\n", indep, indep, res->data[2] & 4 ? "" : " not");
			printf("# %s%sDMA may%s execute in count by byte mode\n", indep, indep, res->data[1] & 8 ? "" : " not");
			printf("# %s%sDMA may%s execute in count by word mode\n", indep, indep, res->data[1] & 0x10 ? "" : " not");
			if ((res->data[1] & 0x60) == 0x00)
				printf("# %s%sDMA channel speed in compatible mode\n", indep, indep);
			if ((res->data[1] & 0x60) == 0x20)
				printf("# %s%sDMA channel speed type A\n", indep, indep);
			if ((res->data[1] & 0x60) == 0x40)
				printf("# %s%sDMA channel speed type B\n", indep, indep);
			if ((res->data[1] & 0x60) == 0x60)
				printf("# %s%sDMA channel speed type F\n", indep, indep);
			printf("%s (DMA %d (CHANNEL %d))\n", comment_if_not_selected, curdma, firstdma);
			curdma++;
			if (!starteddeps)
				depdma = curdma;
			break;
		}
	case StartDep_TAG:
		{
			if (len > 1)
			{
				lengtherror(len, "StartDep_TAG");
				BREAKORCONTINUE;
			}
			putchar('\n');
			if (!starteddeps)
				printf("# Multiple choice time, choose one only !\n\n");
			if (res->len == 0)
			{
				printf("# %sStart dependent functions: priority acceptable\n", indep);
			}
			else
				switch (res->data[0])
				{
				case 0:
					printf("# %sStart dependent functions: priority preferred\n", indep);
					break;
				case 1:
					printf("# %sStart dependent functions: priority acceptable\n", indep);
					break;
				case 2:
					printf("# %sStart dependent functions: priority functional\n", indep);
					break;
				default:
					printf("# %sStart dependent functions: priority INVALID\n", indep);
					break;
				}
			indep = "      ";
			starteddeps = 1;
			curio = depio;
			curdma = depdma;
			curint = depint;
			curmem = depmem;
			break;
		}
	case EndDep_TAG:
		{
			int i;
			if (len > 0)
			{
				lengtherror(len, "EndDep_TAG");
				BREAKORCONTINUE;
			}
			for (i = 0; i < res->end; i++)
			{
				for_each_resource(res->alternatives[i].resources,
								  res->alternatives[i].len,
								  print_resource,
								  stdout,
								  selected && (res->value == i));
			}
			indep = "    ";
			printf("\n# %sEnd dependent functions\n", indep);
			break;
		}
	case IOport_TAG:
		{
			if (len != 7)
			{
				lengtherror(len, "IOport_TAG");
				printf("# Bad tag length in 0x%02x\n", res->tag);
				BREAKORCONTINUE;
			}
			printf("# %sLogical device decodes %s IO address lines\n", indep, res->data[0] ? "16 bit" : "10 bit");
			printf("# %s%sMinimum IO base address 0x%04lx\n", indep, indep, res->start);
			printf("# %s%sMaximum IO base address 0x%04lx\n", indep, indep, res->end - 1);
			printf("# %s%sIO base alignment %ld bytes\n", indep, indep, res->step);
			printf("# %s%sNumber of IO addresses required: %d\n", indep, indep, res->data[6]);
#if DUMPADDR
			for (i = ((res->data[3] << 8) + res->data[2]); i <= ((res->data[5] << 8) + res->data[4]); i += res->data[5])
			{
				printf("# %s%s0x%04x..0x%04x\n", indep, indep, indep, i, i + res->data[6] - 1);
			}
#endif /* DUMPADDR */
			printf("%s (IO %d (SIZE %d) (BASE 0x%04lx)%s)\n",
				   comment_if_not_selected, curio, res->data[6], res->value, IORangeCheck ? " (CHECK)" : "");
			curio++;
			if (!starteddeps)
				depio = curio;
			break;
		}
	case FixedIO_TAG:
		{
			if (len != 3)
			{
				lengtherror(len, "FixedIO_TAG");
				printf("# Bad tag length in 0x%02x\n", res->tag);
				BREAKORCONTINUE;
			}
			printf("# %sFixed IO base address 0x%04lx\n", indep, res->start);
			printf("# %s%sNumber of IO addresses required: %ld\n", indep, indep, res->size);
			printf("%s (IO %d (SIZE %ld) (BASE 0x%04lx)%s)\n",
				   comment_if_not_selected, curio, res->size, res->start, IORangeCheck ? " (CHECK)" : "");
			curio++;
			if (!starteddeps)
				depio = curio;
			break;
		}
	case MemRange_TAG:
		{
			char width = 'w';

			if (len != 9)
			{
				lengtherror(len, "MemRange_TAG");
				printf("# %sInvalid length for memory range tag 0x%02x\n", indep, res->tag);
				BREAKORCONTINUE;
			}
			printf("# %sMemory is %s\n", indep, res->data[0] & 1 ? "writeable" : "non-writeable (ROM)");
			printf("# %sMemory is %s\n", indep, res->data[0] & 2 ? "read cacheable, write-through" : "non-cacheable");
			printf("# %sMemory decode supports %s\n", indep, res->data[0] & 4 ? "range length" : "high address");
			if ((res->data[0] & 0x18) == 0x00)
			{
				width = 'b';
				printf("# %smemory is 8-bit only\n", indep);
			}
			if ((res->data[0] & 0x18) == 0x08)
				printf("# %smemory is 16-bit only\n", indep);
			if ((res->data[0] & 0x18) == 0x10)
				printf("# %smemory is 8-bit and 16-bit\n", indep);
			if (res->data[0] & 0x20)
				printf("# %smemory is shadowable\n", indep);
			if (res->data[0] & 0x40)
				printf("# %smemory is an expansion ROM\n", indep);
			printf("# %sMinimum memory base address 0x%06lx\n", indep, res->start);
			printf("# %sMaximum memory base address 0x%06lx\n", indep, res->end - 1);
			printf("# %sRange base alignment mask 0xff%04lx bytes\n", indep, res->step);
			printf("# %sRange length %ld bytes\n", indep, res->size);
#ifdef DUMPADDR
#if 0
			***untested ***
				for (i = ((res->data[2] << 16) + (res->data[1] << 8));
					 i <= ((res->data[4] << 16) + (res->data[3] << 8));
					 i += ((res->data[6] << 8) + res->data[5]))
			{
				printf("# %s%s0x%06x..0x%06x\n", indep, indep,
					 i, i + (res->data[8] << 16) + (res->data[7] << 8) - 1);
			}
#endif
#endif /* DUMPADDR */
			printf("# Choose UPPER = Range, or UPPER = Upper limit to suit hardware\n");
			printf("%s (MEM %d (BASE 0x%06lx) (MODE %cu) (UPPER 0x%06lx))\n",
				   comment_if_not_selected,
				   curmem, res->start, width, res->start + res->size);
			printf("# (MEM %d (BASE 0x%06lx) (MODE %cr) (UPPER 0x%06lx))\n", curmem, res->start, width, res->size);	/* XXX AJR */
			curmem++;
			if (!starteddeps)
				depmem = curmem;
			break;
		}
	case ANSIstr_TAG:
		{
			printf("# %sANSI string -->", indep);
			/* Remember the device name */
			if(devstring)
				free(devstring);
			devstring = (char *)malloc(len + 1);
			if(!devstring)
			{
				fprintf(stderr, "Out of memory to store board device string\n");
				exit(1);
			}
			for (i = 0; i < len; i++)
				devstring[i] = res->data[i];
			devstring[i] = 0;
			printf("%s<--\n", devstring);
			break;
		}
	case UNICODEstr_TAG:
		{
			printf("# %sUNICODE string -->", indep);
			/* Remember the device name */
			if(devstring)
				free(devstring);
			devstring = (char *)malloc(len + 1);
			if(!devstring)
			{
				fprintf(stderr, "Out of memory to store board device string\n");
				exit(1);
			}
			for (i = 0; i < len; i++)
				putchar(devstring[i] = res->data[i]);
			devstring[i] = 0;
			printf("<--\n");
			break;
		}
	case VendorShort_TAG:
	case VendorLong_TAG:
		{
			printf("# %sVendor defined tag: ", indep);
			printf(" %02x", res->tag);
			for (i = 0; i < len; i++)
				printf(" %02x", res->data[i]);
			putchar('\n');
			break;
		}
	case End_TAG:
		{
			char *indep = "";
			if (nestdepth)
			{
				/* If we have a device name, show it (the last description string before the END DF flag is it) */
				if (serialno == 0xffffffffUL)
					printf(" (NAME \"%s/-1[%d]", devid, logdevno-1);
				else
					printf(" (NAME \"%s/%lu[%d]", devid, serialno, logdevno-1);
				if(devstring)
					printf("{%-20s}", devstring);
				printf("\")\n");
				printf("%s (ACT Y)\n", comment_if_not_selected);
				while (nestdepth)
				{
					printf(")");
					nestdepth--;
				}
				printf("\n");
			}
			/* Cancel any device names */
			if(devstring)
				free(devstring);
			devstring = 0;
			printf("# %sEnd tag... Checksum 0x%02x (%s)\n\n", indep, csum, (csum % 256) ? "BAD" : "OK");
			break;
		}
	case Mem32Range_TAG:
	case FixedMem32Range_TAG:
		{
			if (len != 17)
			{
				lengtherror(len, "Mem32Range_TAG or FixedMem32Range_TAG");
				BREAKORCONTINUE;
			}
			printf("# %s32-bit MemRange tag %02x ...\n", indep, type);
			break;
		}
	default:
		{
			printf("# %sUnknown tag %02x ...\n", indep, type);
			BREAKORCONTINUE;
		}
	}
	return;
#ifdef ABORT_ONRESERR
  oncorruptresourcedata:
	printf("# Resource data dump aborted\n");
	exit(1);
#endif
}

#ifndef __DJGPP__
static void
print_script_header(FILE * output_file)
{
	fprintf(output_file,
			"#!/bin/sh\n"
			"prefixes=\"/etc/pnp/config-scripts/isa /usr/share/pnp/config-scripts/isa\"\n"
			"\n"
			"function pnp_isa_configure {\n"
			"    local i id dir found_it\n"
			"    i=0\n"
			"    while [ \"${dev_ids[$i]}\" != \"\" ] ; do\n"
			"        id=\"${dev_ids[$i]}\"\n"
			"        for dir in $prefixes ; do\n"
			"	    if [ -r $dir/$id ] ; then\n"
			"	        source $dir/$id\n"
			"		return $?\n"
			"	    fi\n"
			"	done\n"
			"	i=`expr $i + 1`\n"
			"    done\n"
			"    # Failed to find anything\n"
			"    return 1\n"
			"}\n"
	);
}

static void
print_script(FILE * output_file, struct resource *res, int selected)
{
	switch (res->type)
	{
	case NewBoard_PSEUDOTAG:
		{
			int csn = res->start;

			serialno = (unsigned long)
				serial_identifier[csn][4] +
				(serial_identifier[csn][5] << 8) +
				(serial_identifier[csn][6] << 16) +
				(serial_identifier[csn][7] << 24);
			strncpy(devid, devidstr(serial_identifier[csn][0],
									serial_identifier[csn][1],
									serial_identifier[csn][2],
									serial_identifier[csn][3]), 8);
		}
		break;
	case PnPVerNo_TAG:
		break;
	case LogDevId_TAG:
		{
			if (nestdepth)
			{
				fprintf(output_file, "pnp_isa_configure\n\n");
				nestdepth = 0;
			}
			fprintf(output_file,
			  "unset irq irq_flags dma dma_flags io_start io_len io_flags\n"
					"unset mem_start mem_len mem_flags dev_ids\n"
					"board_id=%s\n"
					"dev_ids[0]=%s\n",
					devid,
					devidstr(res->data[0], res->data[1],
							 res->data[2], res->data[3]));
			nestdepth = 2;
			curdma = 0;
			depdma = 0;
			curio = 0;
			depio = 0;
			curint = 0;
			depint = 0;
			curmem = 0;
			depmem = 0;
			curid = 2;
			starteddeps = 0;
			break;
		}
	case CompatDevId_TAG:
		{
			fprintf(output_file,
					"dev_ids[%d]=%s\n",
					curid,
					devidstr(res->data[0], res->data[1],
							 res->data[2], res->data[3]));
			curid++;
			break;
		}
	case IRQ_TAG:
		{
			fprintf(output_file,
					"irq[%d]=%ld ; irq_flags[%d]=0x%02x\n",
					curint, res->value, curint, res->data[2]);
			curint++;
			if (!starteddeps)
				depint = curint;
			break;
		}
	case DMA_TAG:
		{
			fprintf(output_file,
					"dma[%d]=%ld ; dma_flags[%d]=0x%02x\n",
					curdma, res->value, curdma, res->data[1]);
			curdma++;
			if (!starteddeps)
				depdma = curdma;
			break;
		}
	case StartDep_TAG:
		{
			starteddeps = 1;
			curio = depio;
			curdma = depdma;
			curint = depint;
			curmem = depmem;
			break;
		}
	case EndDep_TAG:
		{
			for_each_resource(res->alternatives[res->value].resources,
							  res->alternatives[res->value].len,
							  print_script,
							  output_file, 1);
			break;
		}
	case IOport_TAG:
		{
			fprintf(output_file,
					"io_start[%d]=0x%04lx ; "
					"io_len[%d]=%-2ld ; "
					"io_flags[%d]=0x%02x\n",
					curio, res->value,
					curio, res->size,
					curio, res->data[0]);
			curio++;
			if (!starteddeps)
				depio = curio;
			break;
		}
	case FixedIO_TAG:
		{
			fprintf(output_file,
					"io_start[%d]=0x%04lx ; io_len[%d]=%ld ; "
					"io_flags[%d]=FIXED_IO\n",
					curio, res->start, curio, res->size, curio);
			curio++;
			if (!starteddeps)
				depio = curio;
			break;
		}
	case MemRange_TAG:
		{
			fprintf(output_file,
					"mem_start[%d]=0x%04lx ; mem_len[%d]=%ld\n ; "
					"mem_flags[%d]=0x%04x\n",
					curmem, res->value, curmem, res->size,
					curmem, res->data[0]);
			curmem++;
			if (!starteddeps)
				depmem = curmem;
			break;
		}
	case ANSIstr_TAG:
		break;
	case UNICODEstr_TAG:
		break;
	case VendorShort_TAG:
	case VendorLong_TAG:
		break;
	case End_TAG:
		{
			fprintf(output_file, "pnp_isa_configure\n\n");
			break;
		}
	case Mem32Range_TAG:
		{
			/* STUB: Need to add support for this. */
			break;
		}
	case FixedMem32Range_TAG:
		{
			/* STUB: Need to add support for this. */
			break;
		}
	default:
		{
			fprintf(output_file, "# %sUnknown tag %02x ...\n", indep, res->type);
			BREAKORCONTINUE;
		}
	}
	return;
#ifdef ABORT_ONRESERR
  oncorruptresourcedata:
	fprintf(output_file, "# Resource data dump aborted\n");
	exit(1);
#endif
}
#endif /* __DJGPP__ */

static void 
for_each_resource(struct resource *res,
				  int num_resources,
				  resource_handler handler,
				  FILE * output_file,
				  int selected)
{
	int i;
	for (i = 0; i < num_resources; i++)
	{
		(*handler) (output_file, &res[i], selected);
	}
}

static void
read_board(int csn, struct resource **res_list, int *list_len)
{
	int i;
	struct resource *res_array;
	int array_len;

	res_array = *res_list;
	array_len = *list_len;
	if (res_array == NULL || array_len == 0)
	{
		res_array = malloc(sizeof(struct resource));
		array_len = 1;
	}
	else
	{
		array_len++;
		res_array = realloc(res_array,
							array_len * sizeof(struct resource));
	}
	if (!res_array)
	{
		fprintf(stderr,
				"read_board: failed to allocate memory for res_array.\n");
		exit(1);
	}

	res_array[array_len - 1].type = NewBoard_PSEUDOTAG;
	res_array[array_len - 1].start = csn;
	res_array[array_len - 1].end = csn + 1;
	if ((res_array[array_len - 1].data = malloc(IDENT_LEN)) == NULL)
	{
		fprintf(stderr,
		   "read_board: failed to allocate memory for board identifier.\n");
		exit(1);
	}

	Wake(csn);
	/*
	 * Check for broken cards that don't reset their resource pointer
	 * properly
	 */
	/* Get the first byte */
	if (read_resource_data(&res_array[array_len - 1].data[0]) != 0)
		return;
	/* Just check the first byte, if these match, or the serial identifier had a checksum error, we assume it's ok */
	if((!csumdiff[csn])&&(res_array[array_len - 1].data[0] != serial_identifier[csn][0]))
	{
		/*
		 * Assume the card is broken and this is the start of the resource
		 * data.
		 */
		unread_resource_data(res_array[array_len - 1].data[0]);
		res_array[array_len - 1].len = 1;
		goto broken;			/* Wouldn't need this if people read the spec */
	}
	res_array[array_len - 1].len = IDENT_LEN;
	/*
	 * Read resource data past serial identifier - As it should be spec
	 * section 4.5 para 2
	 */
	for (i = 1; i < IDENT_LEN; i++)
	{
		if (read_resource_data(&res_array[array_len - 1].data[i]) != 0)
			return;
	}
	if(csumdiff[csn])
	{
		/* Original serial identifier had checksum error, so make the serial
		 * identifier what we read from the resource data, as that's what
		 * isapnp will see from an (identify *), assuming checksum ok
		 */
		int data;
		int newbit;
		int j;
		csum = 0x6a;
		for (i = 0; i < IDENT_LEN - 1; i++)
		{
			data = res_array[array_len - 1].data[i];
			for(j = 0; j < 8; j++)
			{
				newbit = (data & 1) << 7;
				data >>= 1;
				if(((csum >> 1) ^ csum) & 1)
					newbit ^= 0x80;
				csum >>= 1;
				csum |= newbit;
			}
		}
		if(csum == res_array[array_len - 1].data[IDENT_LEN - 1])
		{
			printf("# WARNING: Making Board %d serial identifier the resource data version\n", csn);
			for (i = 0; i < IDENT_LEN; i++)
			{
				serial_identifier[csn][i] = res_array[array_len - 1].data[i];
			}
		}
		else
			printf("# WARNING: Board %d resource data identifier has checksum error too\n", csn);
	}
	/* Now for the actual resource data */
	csum = 0;
  broken:
	*res_list = res_array;
	*list_len = array_len;
	read_board_resources(NULL, 0, res_list, list_len);
}

#if 0
void
dumpdata(int csn)
{
	int i;
	/* #define OLD_READ_STRATEGY */
#ifdef OLD_READ_STRATEGY
	struct resource res;
#endif /* OLD_READ_STRATEGY */
	unsigned long serialno = (unsigned long) serial_identifier[csn][4] + (serial_identifier[csn][5] << 8) +
	(serial_identifier[csn][6] << 16) + (serial_identifier[csn][7] << 24);
	struct resource *res_array;
	int array_len;
	int alloc_result;

	strncpy(devid, devidstr(serial_identifier[csn][0],
							serial_identifier[csn][1],
							serial_identifier[csn][2],
							serial_identifier[csn][3]), 8);

	nestdepth = 0;
	curdma = 0;
	depdma = 0;
	curio = 0;
	depio = 0;
	curint = 0;
	depint = 0;
	curmem = 0;
	depmem = 0;
	starteddeps = 0;

	Wake(csn);
	printf("# Card %d: (serial identifier", csn);
	for (i = IDENT_LEN; i--;)
		printf(" %02x", serial_identifier[csn][i]);
	printf(")\n");
	if (serialno == 0xffffffffUL)
	{
		printf("# Vendor Id %s, No Serial Number (-1), checksum 0x%02X.\n", devid, serial_identifier[csn][8]);
	}
	else
	{
		printf("# Vendor Id %s, Serial Number %lu, checksum 0x%02X.\n", devid, serialno, serial_identifier[csn][8]);
	}
	/*
	 * Check for broken cards that don't reset their resource pointer
	 * properly
	 */
	for (i = 0; i < TMP_LEN; i++)
		tmp[i] = 0;
	/* Get the first byte */
	if (read_resource_data(&tmp[0]) != 0)
		return;
	/* Just check the first byte, if these match we assume it's ok */
	if (tmp[0] != serial_identifier[csn][0])
	{
		printf("# Ident byte %d, (%02x) differs from resource data (%02x)\n", i, serial_identifier[csn][i], tmp[0]);
		printf("#Assuming the card is broken and this is the start of the resource data\n");
		unread_resource_data(tmp[0]);
		goto broken;			/* Wouldn't need this if people read the spec */
	}
	/*
	 * Read resource data past serial identifier - As it should be spec
	 * section 4.5 para 2
	 */
	for (i = 1; i < IDENT_LEN; i++)
	{
		/* Use tmp[0] as a temporary vbl */
		if (read_resource_data(&tmp[0]) != 0)
			return;
		/*
		 * Checksum byte is not guaranteed, but the previous bytes should
		 * match
		 */
		if ((tmp[0] != serial_identifier[csn][i]) && (i < (IDENT_LEN - 1)))
			printf("# Ident byte %d, (%02x) differs from resource data (%02x)\n", i, serial_identifier[csn][i], tmp[0]);
	}
	/* Now for the actual resource data */
	csum = 0;
  broken:
	res_array = NULL;
	array_len = 0;
	read_board_resources(NULL, 0, &res_array, &array_len);
	if (do_autoconfig)
	{
		alloc_result = alloc_resources(res_array, array_len, NULL, 0);
	}
	else
	{
		alloc_result = 0;
	}
	for_each_resource(res_array, array_len,
					  print_resource, stdout, alloc_result);
}
#endif

void
dumpregs(int csn)
{
	int logical_device;
	int i;
	int addr;
	int desc;
	unsigned long serialno = (unsigned long) serial_identifier[csn][4] + (serial_identifier[csn][5] << 8) +
	(serial_identifier[csn][6] << 16) + (serial_identifier[csn][7] << 24);
	strncpy(devid, devidstr(serial_identifier[csn][0],
							serial_identifier[csn][1],
							serial_identifier[csn][2],
							serial_identifier[csn][3]), 8);

	Wake(csn);
	printf("# Configuration registers for card %d: (serial identifier", csn);
	for (i = IDENT_LEN; i--;)
		printf(" %02x", serial_identifier[csn][i]);
	printf(")\n");
	if (serialno == 0xffffffffUL)
	{
		printf("# Vendor Id %s, No Serial Number (-1), checksum 0x%02X.\n", devid, serial_identifier[csn][8]);
	}
	else
	{
		printf("# Vendor Id %s, Serial Number %lu, checksum 0x%02X.\n", devid, serialno, serial_identifier[csn][8]);
	}
	printf("#\n# Vendor defined card level registers 0x20..2f:");
	for (addr = 0x20; addr < 0x30; addr++)
	{
		ADDRESS(addr);
		printf(" %02x", READ_DATA);
	}
	putchar('\n');
	for (logical_device = 0; logical_device < 256; logical_device++)
	{
		LOGICALDEVICENUMBER;
		WRITE_DATA(logical_device);
		if (READ_DATA != logical_device)
			break;
		printf("#\n# Logical device %d\n", logical_device);
		for (addr = 0x40, desc = 0; addr < 0x60; addr += 8, desc++)
		{
			printf("# 24 bit Memory descriptor %d at %02x..%02x: ", desc, addr, addr + 4);
			for (i = 0; i < 5; i++)
			{
				ADDRESS(addr + i);
				tmp[i] = READ_DATA;
			}
			printf("Base address 0x%02x%02x00 %s 0x%02x%02x00, %d bit\n", tmp[0], tmp[1],
				   tmp[2] & 1 ? "...." : "size",
				   tmp[3], tmp[4], tmp[2] & 2 ? 16 : 8);
		}
		for (addr = 0x76, desc = 0; addr < 0xb0; addr += 16, desc++)
		{
			printf("# 32 bit Memory descriptor %d at %02x..%02x: ", desc, addr, addr + 8);
			for (i = 0; i < 9; i++)
			{
				ADDRESS(addr + i);
				tmp[i] = READ_DATA;
			}
			printf("Base address 0x%02x%02x%02x%02x %s 0x%02x%02x%02x%02x, %d bit\n", tmp[0], tmp[1], tmp[2], tmp[3],
				   tmp[4] & 1 ? "...." : "size",
				   tmp[5], tmp[6], tmp[7], tmp[8], tmp[8] & 4 ? 32 : (tmp[8] & 2 ? 16 : 8));
			addr &= 0xf0;
		}
		for (addr = 0x60, desc = 0; addr < 0x70; addr += 2, desc++)
		{
			printf("# IO descriptor %d at %02x..%02x: ", desc, addr, addr + 1);
			for (i = 0; i < 2; i++)
			{
				ADDRESS(addr + i);
				tmp[i] = READ_DATA;
			}
			printf("Base address 0x%02x%02x\n", tmp[0], tmp[1]);
		}
		for (addr = 0x70, desc = 0; addr < 0x74; addr += 2, desc++)
		{
			printf("# Interrupt descriptor %d at %02x..%02x: ", desc, addr, addr + 1);
			for (i = 0; i < 2; i++)
			{
				ADDRESS(addr + i);
				tmp[i] = READ_DATA;
			}
			printf("Interrupt level %d, active %s, %s triggered\n", tmp[0], tmp[1] & 2 ? "high" : "low", tmp[1] & 1 ? "level" : "edge");
		}
		for (addr = 0x74; addr < 0x76; addr++)
		{
			printf("# DMA descriptor %d at %02x: ", addr - 0x74, addr);
			ADDRESS(addr);
			tmp[0] = READ_DATA;
			printf("DMA channel %d\n", tmp[0]);
		}
	}
	putchar('\n');
}
